diff -uNr trunk/common/common.c trunk-new/common/common.c
--- trunk/common/common.c	Tue Jul 17 08:11:18 2007
+++ trunk-new/common/common.c	Fri Jan  4 19:45:58 2008
@@ -245,6 +245,13 @@
         else
             p->i_threads = atoi(value);
     }
+    OPT("thread-queue")
+    {
+        if( !strcmp(value, "auto") )
+            p->i_thread_queue = 0;
+        else
+            p->i_thread_queue = atoi(value);
+    }
     OPT2("deterministic", "n-deterministic")
         p->b_deterministic = atobool(value);
     OPT2("level", "level-idc")
diff -uNr trunk/common/common.h trunk-new/common/common.h
--- trunk/common/common.h	Thu Dec 20 16:24:16 2007
+++ trunk-new/common/common.h	Fri Jan  4 19:45:58 2008
@@ -233,12 +233,19 @@
 struct x264_t
 {
     /* encoder parameters */
-    x264_param_t    param;
+    x264_param_t         param;
 
-    x264_t          *thread[X264_THREAD_MAX];
-    x264_pthread_t  thread_handle;
-    int             b_thread_active;
-    int             i_thread_phase; /* which thread to use for the next frame */
+    x264_t               *thread[X264_THREAD_MAX]; /* contexts for each frame in progress */
+    x264_t               **thread_queue; /* frames that have been prepared but not yet claimed by a worker thread */
+    x264_pthread_cond_t  thread_queue_cv;
+    x264_pthread_mutex_t thread_queue_mutex;
+    x264_pthread_t       *thread_handle;
+    int                  thread_active;
+    x264_pthread_cond_t  thread_active_cv;
+    x264_pthread_mutex_t thread_active_mutex;
+    int                  thread_exit;
+    int                  b_thread_active;
+    int                  i_thread_phase; /* which thread to use for the next frame */
 
     /* bitstream output */
     struct
diff -uNr trunk/common/csp.c trunk-new/common/csp.c
--- trunk/common/csp.c	Tue Jul 17 08:11:18 2007
+++ trunk-new/common/csp.c	Fri Jan  4 19:46:36 2008
@@ -64,7 +64,7 @@
         int     i;
         for( i = 0; i < w; i++ )
         {
-            *d++ = ( s[0] + s[1] + s[i_src] + s[i_src+1] + 1 ) >> 2;
+            *d++ = ( s[0] + s[1] + s[i_src] + s[i_src+1] + 2 ) >> 2;
             s += 2;
         }
         dst += i_dst;
@@ -254,24 +254,64 @@
     }
 }
 
-/* Same value than in XviD */
-#define BITS 8
-#define FIX(f) ((int)((f) * (1 << BITS) + 0.5))
-
-#define Y_R   FIX(0.257)
-#define Y_G   FIX(0.504)
-#define Y_B   FIX(0.098)
-#define Y_ADD 16
-
-#define U_R   FIX(0.148)
-#define U_G   FIX(0.291)
-#define U_B   FIX(0.439)
-#define U_ADD 128
-
-#define V_R   FIX(0.439)
-#define V_G   FIX(0.368)
-#define V_B   FIX(0.071)
-#define V_ADD 128
+#ifdef BT_709
+//BT.709
+#define Kb 0.0722
+#define Kr 0.2126
+#else
+//BT.601
+#define Kb 0.114
+#define Kr 0.299
+#endif
+
+#define Kg (1.0 - Kb - Kr)
+#define Sb (1.0 - Kb)
+#define Sr (1.0 - Kr)
+
+#ifdef PC_SCALE
+//PC Scale
+//0 <= Y <= 255
+//0 <= U <= 255
+//0 <= V <= 255
+#define Ky 1.0
+#define Ku (0.5 / Sb)
+#define Kv (0.5 / Sr)
+#define Ay 0.0
+#define Au 127.5
+#define Av 127.5
+#else
+//TV Scale
+//16 <= Y <= 235
+//16 <= U <= 240
+//16 <= V <= 240
+#define Ky (1.0        * 219.0 / 255.0)
+#define Ku ((0.5 / Sb) * 224.0 / 255.0)
+#define Kv ((0.5 / Sr) * 224.0 / 255.0)
+#define Ay 16.0
+#define Au 128.0
+#define Av 128.0
+#endif
+
+#define BITS      22
+#define INT_FIX   (1 << BITS)
+#define INT_ROUND (INT_FIX >> 1)
+#define FIX(f)    ((uint32_t)((f) * INT_FIX + 0.5))
+
+#define Y_R   FIX(Kr*Ky)
+#define Y_G   FIX(Kg*Ky)
+#define Y_B   FIX(Kb*Ky)
+#define Y_ADD ((uint32_t)(Ay * INT_FIX + INT_ROUND + 0.5))
+
+#define U_R   FIX(Kr*Ku)
+#define U_G   FIX(Kg*Ku)
+#define U_B   FIX(Sb*Ku)
+#define U_ADD ((uint32_t)((Au * INT_FIX + INT_ROUND) * 4 + 0.5))
+
+#define V_R   FIX(Sr*Kv)
+#define V_G   FIX(Kg*Kv)
+#define V_B   FIX(Kb*Kv)
+#define V_ADD ((uint32_t)((Av * INT_FIX + INT_ROUND) * 4 + 0.5))
+
 #define RGB_TO_I420( name, POS_R, POS_G, POS_B, S_RGB ) \
 static void name( x264_mc_functions_t *mc,              \
                   x264_frame_t *frm, x264_image_t *img, \
@@ -300,20 +340,20 @@
                                                         \
         for( w = i_width; w > 0; w -= 2 )               \
         {                                               \
-            int cr = 0,cg = 0,cb = 0;                   \
-            int r, g, b;                                \
+            uint32_t cr = 0,cg = 0,cb = 0;              \
+            uint32_t r, g, b;                           \
                                                         \
             /* Luma */                                  \
             cr = r = ss[POS_R];                         \
             cg = g = ss[POS_G];                         \
             cb = b = ss[POS_B];                         \
                                                         \
-            yy[0] = Y_ADD + ((Y_R * r + Y_G * g + Y_B * b) >> BITS);    \
+            yy[0] = (Y_ADD + Y_R * r + Y_G * g + Y_B * b) >> BITS;    \
                                                         \
             cr+= r = ss[POS_R+i_src];                   \
             cg+= g = ss[POS_G+i_src];                   \
             cb+= b = ss[POS_B+i_src];                   \
-            yy[i_y] = Y_ADD + ((Y_R * r + Y_G * g + Y_B * b) >> BITS);  \
+            yy[i_y] = (Y_ADD + Y_R * r + Y_G * g + Y_B * b) >> BITS;  \
             yy++;                                       \
             ss += S_RGB;                                \
                                                         \
@@ -321,21 +361,21 @@
             cg+= g = ss[POS_G];                         \
             cb+= b = ss[POS_B];                         \
                                                         \
-            yy[0] = Y_ADD + ((Y_R * r + Y_G * g + Y_B * b) >> BITS);    \
+            yy[0] = (Y_ADD + Y_R * r + Y_G * g + Y_B * b) >> BITS;    \
                                                         \
             cr+= r = ss[POS_R+i_src];                   \
             cg+= g = ss[POS_G+i_src];                   \
             cb+= b = ss[POS_B+i_src];                   \
-            yy[i_y] = Y_ADD + ((Y_R * r + Y_G * g + Y_B * b) >> BITS);  \
+            yy[i_y] = (Y_ADD + Y_R * r + Y_G * g + Y_B * b) >> BITS;  \
             yy++;                                       \
             ss += S_RGB;                                \
                                                         \
             /* Chroma */                                \
-            *uu++ = (uint8_t)(U_ADD + ((-U_R * cr - U_G * cg + U_B * cb) >> (BITS+2)) ); \
-            *vv++ = (uint8_t)(V_ADD + (( V_R * cr - V_G * cg - V_B * cb) >> (BITS+2)) ); \
+            *uu++ = (uint8_t)((U_ADD + U_B * cb - U_R * cr - U_G * cg) >> (BITS+2)); \
+            *vv++ = (uint8_t)((V_ADD + V_R * cr - V_G * cg - V_B * cb) >> (BITS+2)); \
         }                                               \
                                                         \
-        src += 2*i_src;                                   \
+        src += 2*i_src;                                 \
         y += 2*frm->i_stride[0];                        \
         u += frm->i_stride[1];                          \
         v += frm->i_stride[2];                          \
diff -uNr trunk/configure trunk-new/configure
--- trunk/configure	Sun Dec  2 22:16:56 2007
+++ trunk-new/configure	Fri Jan  4 19:46:48 2008
@@ -405,9 +405,9 @@
 fi
 
 if [ "$debug" = "yes" ]; then
-    CFLAGS="-O1 -g $CFLAGS"
+    CFLAGS="-O1 -g -D_DEBUG $CFLAGS"
 else
-    CFLAGS="-O4 -ffast-math $CFLAGS"
+    CFLAGS="-O4 -ffast-math -DNDEBUG $CFLAGS"
 fi
 
 if cc_check "stdio.h" "" "fseeko(stdin,0,0);" ; then
diff -uNr trunk/encoder/encoder.c trunk-new/encoder/encoder.c
--- trunk/encoder/encoder.c	Tue Nov 20 02:57:28 2007
+++ trunk-new/encoder/encoder.c	Fri Jan  4 19:46:48 2008
@@ -63,6 +63,34 @@
                                     x264_nal_t **pp_nal, int *pi_nal,
                                     x264_picture_t *pic_out );
 
+/* threading */
+
+static int x264_slices_write_thread( x264_t *h );
+
+#ifdef HAVE_PTHREAD
+void x264_cond_broadcast( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{
+    x264_pthread_mutex_lock( mutex );
+    *var = val;
+    x264_pthread_cond_broadcast( cv );
+    x264_pthread_mutex_unlock( mutex );
+}
+
+void x264_cond_wait( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{
+    x264_pthread_mutex_lock( mutex );
+    while( *var != val )
+        x264_pthread_cond_wait( cv, mutex );
+    x264_pthread_mutex_unlock( mutex );
+}
+
+#else
+void x264_cond_broadcast( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{}
+void x264_cond_wait( x264_pthread_cond_t *cv, x264_pthread_mutex_t *mutex, int *var, int val )
+{}
+#endif
+
 /****************************************************************************
  *
  ******************************* x264 libs **********************************
@@ -359,7 +387,12 @@
         if( h->param.i_scenecut_threshold >= 0 )
             h->param.b_pre_scenecut = 1;
 #endif
+        if( h->param.i_thread_queue == 0 )
+            h->param.i_thread_queue = h->param.i_threads;
+        h->param.i_thread_queue = x264_clip3( h->param.i_thread_queue, h->param.i_threads, X264_THREAD_MAX );
     }
+    else
+        h->param.i_thread_queue = 1;
 
     if( h->param.b_interlaced )
     {
@@ -615,7 +648,7 @@
     x264_reduce_fraction( &h->param.i_fps_num, &h->param.i_fps_den );
 
     /* Init x264_t */
-    h->i_frame = 0;
+    h->i_frame = -1;
     h->i_frame_num = 0;
     h->i_idr_pic_id = 0;
 
@@ -636,7 +669,7 @@
     h->mb.i_mb_count = h->sps->i_mb_width * h->sps->i_mb_height;
 
     /* Init frames. */
-    h->frames.i_delay = h->param.i_bframe + h->param.i_threads - 1;
+    h->frames.i_delay = h->param.i_bframe + h->param.i_thread_queue - 1;
     h->frames.i_max_ref0 = h->param.i_frame_reference;
     h->frames.i_max_ref1 = h->sps->vui.i_num_reorder_frames;
     h->frames.i_max_dpb  = h->sps->vui.i_max_dec_frame_buffering;
@@ -693,17 +726,39 @@
 
     h->thread[0] = h;
     h->i_thread_num = 0;
-    for( i = 1; i < h->param.i_threads; i++ )
+    for( i = 1; i < h->param.i_thread_queue; i++ )
         h->thread[i] = x264_malloc( sizeof(x264_t) );
 
-    for( i = 0; i < h->param.i_threads; i++ )
+    if( h->param.i_threads > 1 )
     {
+        h->thread_handle = x264_malloc( h->param.i_threads * sizeof(x264_pthread_t) );
+        h->thread_queue = x264_malloc( (h->param.i_thread_queue + 1) * sizeof(x264_t*) );
+        memset( h->thread_queue, 0, (h->param.i_thread_queue + 1) * sizeof(x264_t*) );
+        x264_pthread_cond_init( &h->thread_queue_cv, NULL );
+        x264_pthread_mutex_init( &h->thread_queue_mutex, NULL );
+    }
+
+    for( i = 0; i < h->param.i_thread_queue; i++ )
+    {
+        x264_t *t = h->thread[i];
         if( i > 0 )
-            *h->thread[i] = *h;
-        h->thread[i]->fdec = x264_frame_pop_unused( h );
-        h->thread[i]->out.p_bitstream = x264_malloc( h->out.i_bitstream );
-        if( x264_macroblock_cache_init( h->thread[i] ) < 0 )
+            *t = *h;
+        t->fdec = x264_frame_pop_unused( h );
+        t->out.p_bitstream = x264_malloc( h->out.i_bitstream );
+        if( x264_macroblock_cache_init( t ) < 0 )
             return NULL;
+
+        if( h->param.i_threads > 1 )
+        {
+            x264_pthread_cond_init( &t->thread_active_cv, NULL );
+            x264_pthread_mutex_init( &t->thread_active_mutex, NULL );
+        }
+    }
+
+    if( h->param.i_threads > 1 )
+    {
+        for( i = 0; i < h->param.i_threads; i++ )
+            x264_pthread_create( &h->thread_handle[i], NULL, (void*)x264_slices_write_thread, h );
     }
 
     if( x264_ratecontrol_new( h ) < 0 )
@@ -942,9 +997,6 @@
 {
     int i;
 
-    if( h->fdec->i_frame >= 0 )
-        h->i_frame++;
-
     if( !h->fdec->b_kept_as_ref )
     {
         if( h->param.i_threads > 1 )
@@ -1234,6 +1286,36 @@
     return 0;
 }
 
+static int x264_slices_write_thread( x264_t *h )
+{
+    for(;;)
+    {
+        int i_frame_size;
+        x264_t *t = NULL;
+
+        // get one frame from the queue
+        x264_pthread_mutex_lock( &h->thread_queue_mutex );
+        while( !h->thread_queue[0] && !h->thread_exit )
+            x264_pthread_cond_wait( &h->thread_queue_cv, &h->thread_queue_mutex );
+        if( h->thread_queue[0] )
+            t = (void*)x264_frame_shift( (void*)h->thread_queue );
+        x264_pthread_mutex_unlock( &h->thread_queue_mutex );
+        if( h->thread_exit )
+            return 0;
+        if( !t )
+            continue;
+
+        x264_stack_align( x264_slice_write, t );
+        i_frame_size = t->out.nal[t->out.i_nal-1].i_payload;
+        x264_fdec_filter_row( t, t->sps->i_mb_height );
+        t->out.i_frame_size = i_frame_size;
+
+        x264_cond_broadcast( &t->thread_active_cv, &t->thread_active_mutex, &t->thread_active, 0 );
+    }
+
+    return 0;
+}
+
 /****************************************************************************
  * x264_encoder_encode:
  *  XXX: i_poc   : is the poc of the current given picture
@@ -1261,7 +1343,7 @@
     if( h->param.i_threads > 1)
     {
         int i = ++h->i_thread_phase;
-        int t = h->param.i_threads;
+        int t = h->param.i_thread_queue;
         thread_current = h->thread[ i%t ];
         thread_prev    = h->thread[ (i-1)%t ];
         thread_oldest  = h->thread[ (i+1)%t ];
@@ -1305,7 +1387,7 @@
         if( h->frames.b_have_lowres )
             x264_frame_init_lowres( h, fenc );
 
-        if( h->frames.i_input <= h->frames.i_delay + 1 - h->param.i_threads )
+        if( h->frames.i_input <= h->frames.i_delay + 1 - h->param.i_thread_queue )
         {
             /* Nothing yet to encode */
             /* waiting for filling bframe buffer */
@@ -1314,6 +1396,7 @@
         }
     }
 
+    h->i_frame++;
     if( h->frames.current[0] == NULL )
     {
         int bframes = 0;
@@ -1478,8 +1561,14 @@
     /* Write frame */
     if( h->param.i_threads > 1 )
     {
-        x264_pthread_create( &h->thread_handle, NULL, (void*)x264_slices_write, h );
+        assert( h->thread_active == 0 );
+        h->thread_active = 1;
+        assert( h->b_thread_active == 0 );
         h->b_thread_active = 1;
+        x264_pthread_mutex_lock( &h->thread[0]->thread_queue_mutex );
+        x264_frame_push( (void*)h->thread_queue, (void*)h );
+        x264_pthread_cond_broadcast( &h->thread[0]->thread_queue_cv );
+        x264_pthread_mutex_unlock( &h->thread[0]->thread_queue_mutex );
     }
     else
         x264_slices_write( h );
@@ -1598,7 +1687,7 @@
 
     if( h->b_thread_active )
     {
-        x264_pthread_join( h->thread_handle, NULL );
+        x264_cond_wait( &h->thread_active_cv, &h->thread_active_mutex, &h->thread_active, 0 );
         h->b_thread_active = 0;
     }
     if( !h->out.i_nal )
@@ -1769,12 +1858,41 @@
     int64_t i_yuv_size = 3 * h->param.i_width * h->param.i_height / 2;
     int i;
 
-    for( i=0; i<h->param.i_threads; i++ )
+    if( h->param.i_threads > 1 )
     {
         // don't strictly have to wait for the other threads, but it's simpler than cancelling them
-        if( h->thread[i]->b_thread_active )
-            x264_pthread_join( h->thread[i]->thread_handle, NULL );
+        h->thread_exit = 1;
+        x264_pthread_mutex_lock( &h->thread_queue_mutex );
+        x264_pthread_cond_broadcast( &h->thread_queue_cv );
+        x264_pthread_mutex_unlock( &h->thread_queue_mutex );
+        for( i=0; i < h->param.i_threads; i++ )
+            x264_pthread_join( h->thread_handle[i], NULL );
+        for( i=0; i < h->param.i_thread_queue; i++ )
+        {
+            x264_pthread_cond_destroy( &h->thread[i]->thread_active_cv );
+            x264_pthread_mutex_destroy( &h->thread[i]->thread_active_mutex );
+            if( h->thread[i]->b_thread_active )
+            {
+                assert( h->thread[i]->fenc->i_reference_count == 1 );
+                x264_frame_delete( h->thread[i]->fenc );
+            }
+        }
+        x264_pthread_cond_destroy( &h->thread_queue_cv );
+        x264_pthread_mutex_destroy( &h->thread_queue_mutex );
+        x264_free( h->thread_handle );
+        x264_free( h->thread_queue );
+    }
+
+    if( h->param.i_threads > 1)
+    {
+        x264_t *thread_prev;
+
+        thread_prev = h->thread[ h->i_thread_phase % h->param.i_thread_queue ];
+        x264_thread_sync_ratecontrol( h, thread_prev, h );
+        x264_thread_sync_ratecontrol( thread_prev, thread_prev, h );
+        h->i_frame = thread_prev->i_frame + 1 - h->param.i_thread_queue;
     }
+    h->i_frame++;
 
 #ifdef DEBUG_BENCHMARK
     x264_log( h, X264_LOG_INFO,
@@ -1938,16 +2056,6 @@
             x264_log( h, X264_LOG_INFO, "kb/s:%.1f\n", f_bitrate );
     }
 
-    /* frames */
-    for( i = 0; h->frames.current[i]; i++ )
-        x264_frame_delete( h->frames.current[i] );
-    for( i = 0; h->frames.next[i]; i++ )
-        x264_frame_delete( h->frames.next[i] );
-    for( i = 0; h->frames.unused[i]; i++ )
-        x264_frame_delete( h->frames.unused[i] );
-    for( i = 0; h->frames.reference[i]; i++ )
-        x264_frame_delete( h->frames.reference[i] );
-
     /* rc */
     x264_ratecontrol_delete( h );
 
@@ -1960,8 +2068,46 @@
         free( h->param.rc.psz_rc_eq );
 
     x264_cqm_delete( h );
-    for( i = h->param.i_threads - 1; i >= 0; i-- )
+
+    if( h->param.i_threads > 1)
+        h = h->thread[ h->i_thread_phase % h->param.i_thread_queue ];
+
+    /* frames */
+    for( i = 0; h->frames.current[i]; i++ )
+    {
+        assert( h->frames.current[i]->i_reference_count == 1 );
+        x264_frame_delete( h->frames.current[i] );
+    }
+    for( i = 0; h->frames.next[i]; i++ )
+    {
+        assert( h->frames.next[i]->i_reference_count == 1 );
+        x264_frame_delete( h->frames.next[i] );
+    }
+    for( i = 0; h->frames.unused[i]; i++ )
     {
+        assert( h->frames.unused[i]->i_reference_count == 0 );
+        x264_frame_delete( h->frames.unused[i] );
+    }
+
+    h = h->thread[0];
+
+    for( i = h->param.i_thread_queue - 1; i >= 0; i-- )
+    {
+        x264_frame_t **frame;
+
+        for( frame = h->thread[i]->frames.reference; *frame; frame++ )
+        {
+            assert( (*frame)->i_reference_count > 0 );
+            (*frame)->i_reference_count--;
+            if( (*frame)->i_reference_count == 0 )
+                x264_frame_delete( *frame );
+        }
+        frame = &h->thread[i]->fdec;
+        assert( (*frame)->i_reference_count > 0 );
+        (*frame)->i_reference_count--;
+        if( (*frame)->i_reference_count == 0 )
+            x264_frame_delete( *frame );
+
         x264_macroblock_cache_end( h->thread[i] );
         x264_free( h->thread[i]->out.p_bitstream );
         x264_free( h->thread[i] );
diff -uNr trunk/encoder/macroblock.c trunk-new/encoder/macroblock.c
--- trunk/encoder/macroblock.c	Sun Dec  2 12:57:42 2007
+++ trunk-new/encoder/macroblock.c	Fri Jan  4 20:52:40 2008
@@ -598,7 +598,7 @@
  *  Check if the current MB could be encoded as a [PB]_SKIP (it supposes you use
  *  the previous QP
  *****************************************************************************/
-int x264_macroblock_probe_skip( x264_t *h, const int b_bidir )
+int x264_macroblock_probe_skip( x264_t *h, int b_bidir )
 {
     DECLARE_ALIGNED( int16_t, dct4x4[16][4][4], 16 );
     DECLARE_ALIGNED( int16_t, dct2x2[2][2], 16 );
diff -uNr trunk/encoder/ratecontrol.c trunk-new/encoder/ratecontrol.c
--- trunk/encoder/ratecontrol.c	Wed Sep 12 02:42:22 2007
+++ trunk-new/encoder/ratecontrol.c	Fri Jan  4 19:46:48 2008
@@ -176,8 +176,8 @@
 
     x264_cpu_restore( h->param.cpu );
 
-    rc = h->rc = x264_malloc( h->param.i_threads * sizeof(x264_ratecontrol_t) );
-    memset( rc, 0, h->param.i_threads * sizeof(x264_ratecontrol_t) );
+    rc = h->rc = x264_malloc( h->param.i_thread_queue * sizeof(x264_ratecontrol_t) );
+    memset( rc, 0, h->param.i_thread_queue * sizeof(x264_ratecontrol_t) );
 
     rc->b_abr = h->param.rc.i_rc_method != X264_RC_CQP && !h->param.rc.b_stat_read;
     rc->b_2pass = h->param.rc.i_rc_method == X264_RC_ABR && h->param.rc.b_stat_read;
@@ -358,21 +358,19 @@
             x264_log( h, X264_LOG_WARNING, "2nd pass has fewer frames than 1st pass (%d vs %d)\n",
                       h->param.i_frame_total, rc->num_entries );
         }
-        if( h->param.i_frame_total > rc->num_entries + h->param.i_bframe )
+        if( h->param.i_frame_total > rc->num_entries + h->param.i_bframe + X264_THREAD_MAX )
         {
             x264_log( h, X264_LOG_ERROR, "2nd pass has more frames than 1st pass (%d vs %d)\n",
                       h->param.i_frame_total, rc->num_entries );
             return -1;
         }
 
-        /* FIXME: ugly padding because VfW drops delayed B-frames */
-        rc->num_entries += h->param.i_bframe;
-
-        rc->entry = (ratecontrol_entry_t*) x264_malloc(rc->num_entries * sizeof(ratecontrol_entry_t));
-        memset(rc->entry, 0, rc->num_entries * sizeof(ratecontrol_entry_t));
+        /* FIXME: ugly padding because VfW drops delayed B-frames and frames delayed by multithreading */
+        rc->entry = (ratecontrol_entry_t*) x264_malloc((rc->num_entries + h->param.i_bframe + X264_THREAD_MAX) * sizeof(ratecontrol_entry_t));
+        memset(rc->entry, 0, (rc->num_entries + h->param.i_bframe + X264_THREAD_MAX) * sizeof(ratecontrol_entry_t));
 
         /* init all to skipped p frames */
-        for(i=0; i<rc->num_entries; i++){
+        for(i=0; i < rc->num_entries + h->param.i_bframe + X264_THREAD_MAX; i++){
             ratecontrol_entry_t *rce = &rc->entry[i];
             rce->pict_type = SLICE_TYPE_P;
             rce->qscale = rce->new_qscale = qp2qscale(20);
@@ -382,7 +380,7 @@
 
         /* read stats */
         p = stats_in;
-        for(i=0; i < rc->num_entries - h->param.i_bframe; i++){
+        for(i=0; i < rc->num_entries; i++){
             ratecontrol_entry_t *rce;
             int frame_number;
             char pict_type;
@@ -397,7 +395,7 @@
             }
             e = sscanf(p, " in:%d ", &frame_number);
 
-            if(frame_number < 0 || frame_number >= rc->num_entries)
+            if(frame_number < 0 || frame_number >= rc->num_entries + h->param.i_bframe + X264_THREAD_MAX)
             {
                 x264_log(h, X264_LOG_ERROR, "bad frame number (%d) at stats line %d\n", frame_number, i);
                 return -1;
@@ -457,7 +455,7 @@
         x264_free( p );
     }
 
-    for( i=1; i<h->param.i_threads; i++ )
+    for( i=1; i<h->param.i_thread_queue; i++ )
     {
         h->thread[i]->rc = rc+i;
         rc[i] = rc[0];
@@ -602,7 +600,7 @@
     if( rc->p_stat_file_out )
     {
         fclose( rc->p_stat_file_out );
-        if( h->i_frame >= rc->num_entries - h->param.i_bframe )
+        if( h->i_frame >= rc->num_entries )
             if( rename( rc->psz_stat_file_tmpname, h->param.rc.psz_stat_out ) != 0 )
             {
                 x264_log( h, X264_LOG_ERROR, "failed to rename \"%s\" to \"%s\"\n",
@@ -656,7 +654,7 @@
     if( h->param.rc.b_stat_read )
     {
         int frame = h->fenc->i_frame;
-        assert( frame >= 0 && frame < rc->num_entries );
+        assert( frame >= 0 && frame < rc->num_entries + h->param.i_bframe + X264_THREAD_MAX );
         rce = h->rc->rce = &h->rc->entry[frame];
 
         if( h->sh.i_type == SLICE_TYPE_B
@@ -833,7 +831,7 @@
     x264_ratecontrol_t *rc = h->rc;
     if( h->param.rc.b_stat_read )
     {
-        if( frame_num >= rc->num_entries )
+        if( frame_num >= rc->num_entries + h->param.i_bframe + X264_THREAD_MAX )
         {
             /* We could try to initialize everything required for ABR and
              * adaptive B-frames, but that would be complicated.
@@ -1163,9 +1161,9 @@
     {
         int j = h->rc - h->thread[0]->rc;
         int i;
-        for( i=1; i<h->param.i_threads; i++ )
+        for( i=1; i<h->param.i_thread_queue; i++ )
         {
-            x264_t *t = h->thread[ (j+i)%h->param.i_threads ];
+            x264_t *t = h->thread[ (j+i)%h->param.i_thread_queue ];
             double bits = t->rc->frame_size_planned;
             if( !t->b_thread_active )
                 continue;
@@ -1314,10 +1312,15 @@
         if( rcc->b_2pass )
         {
             //FIXME adjust abr_buffer based on distance to the end of the video
-            int64_t diff = total_bits - (int64_t)rce.expected_bits;
+            int64_t diff;
+
+            if( h->fenc->i_frame < h->param.i_thread_queue )
+                diff = total_bits + (int64_t)h->fenc->i_frame * rcc->bitrate / rcc->fps - (int64_t)rce.expected_bits;
+            else
+                diff = total_bits + (int64_t)(h->param.i_thread_queue - 1) * rcc->bitrate / rcc->fps - (int64_t)rce.expected_bits;
             q = rce.new_qscale;
             q /= x264_clip3f((double)(abr_buffer - diff) / abr_buffer, .5, 2);
-            if( h->fenc->i_frame > 30 )
+            if( h->fenc->i_frame + 1 - h->param.i_thread_queue >= rcc->fps && rcc->expected_bits_sum > 0)
             {
                 /* Adjust quant based on the difference between
                  * achieved and expected bitrate so far */
@@ -1363,7 +1366,7 @@
             }
             else
             {
-                int i_frame_done = h->fenc->i_frame + 1 - h->param.i_threads;
+                int i_frame_done = h->fenc->i_frame + 1 - h->param.i_thread_queue;
 
                 q = get_qscale( h, &rce, rcc->wanted_bits_window / rcc->cplxr_sum, h->fenc->i_frame );
 
@@ -1371,7 +1374,7 @@
                 wanted_bits = i_frame_done * rcc->bitrate / rcc->fps;
                 if( wanted_bits > 0 )
                 {
-                    abr_buffer *= X264_MAX( 1, sqrt(i_frame_done/25) );
+                    abr_buffer *= X264_MAX( 1, sqrt(i_frame_done / rcc->fps) );
                     overflow = x264_clip3f( 1.0 + (total_bits - wanted_bits) / abr_buffer, .5, 2 );
                     q *= overflow;
                 }
diff -uNr trunk/encoder/slicetype.c trunk-new/encoder/slicetype.c
--- trunk/encoder/slicetype.c	Mon Nov 19 14:10:56 2007
+++ trunk-new/encoder/slicetype.c	Fri Jan  4 19:46:48 2008
@@ -313,7 +313,7 @@
         x264_cpu_restore( h->param.cpu );
     }
 
-    if( b_intra_penalty )
+    if( b_intra_penalty && h->sps->i_mb_width > 2 && h->sps->i_mb_height > 2 )
     {
         // arbitrary penalty for I-blocks after B-frames
         int nmb = (h->sps->i_mb_width - 2) * (h->sps->i_mb_height - 2);
@@ -352,7 +352,11 @@
     if( res )
     {
         int imb = frame->i_intra_mbs[pdist];
-        int pmb = (h->sps->i_mb_width - 2) * (h->sps->i_mb_height - 2) - imb;
+        int pmb;
+        if( h->sps->i_mb_width > 2 && h->sps->i_mb_height > 2 )
+            pmb = (h->sps->i_mb_width - 2) * (h->sps->i_mb_height - 2) - imb;
+        else
+            pmb = -imb;
         x264_log( h, X264_LOG_DEBUG, "scene cut at %d Icost:%d Pcost:%d ratio:%.4f bias:%.4f gop:%d (imb:%d pmb:%d)\n",
                   frame->i_frame,
                   icost, pcost, 1. - (double)pcost / icost,
@@ -368,10 +372,15 @@
     int num_frames;
     int keyint_limit;
     int j;
-    int i_mb_count = (h->sps->i_mb_width - 2) * (h->sps->i_mb_height - 2);
+    int i_mb_count;
     int cost1p0, cost2p0, cost1b1, cost2p1;
     int idr_frame_type;
 
+    if( h->sps->i_mb_width > 2 && h->sps->i_mb_height > 2 )
+        i_mb_count = (h->sps->i_mb_width - 2) * (h->sps->i_mb_height - 2);
+    else
+        i_mb_count = 0;
+
     assert( h->frames.b_have_lowres );
 
     if( !h->frames.last_nonb )
@@ -421,9 +430,14 @@
     {
         int pthresh = X264_MAX(INTER_THRESH - P_SENS_BIAS * (j-1), INTER_THRESH/10);
         int pcost = x264_slicetype_frame_cost( h, &a, frames, 0, j+1, j+1, 1 );
-//      fprintf( stderr, "frm%d+%d: %d <=> %d, I:%d/%d \n",
-//               frames[0]->i_frame, j-1, pthresh, pcost/i_mb_count,
-//               frames[j+1]->i_intra_mbs[j+1], i_mb_count );
+//      if( i_mb_count )
+//          fprintf( stderr, "frm%d+%d: %d <=> %d, I:%d/%d \n",
+//                   frames[0]->i_frame, j-1, pthresh, pcost/i_mb_count,
+//                   frames[j+1]->i_intra_mbs[j+1], i_mb_count );
+//      else
+//          fprintf( stderr, "frm%d+%d: %d <=> %d, I:%d/%d \n",
+//                   frames[0]->i_frame, j-1, pthresh, pcost,
+//                   frames[j+1]->i_intra_mbs[j+1], i_mb_count );
         if( pcost > pthresh*i_mb_count || frames[j+1]->i_intra_mbs[j+1] > i_mb_count/3 )
         {
             frames[j]->i_type = X264_TYPE_P;
diff -uNr trunk/x264.c trunk-new/x264.c
--- trunk/x264.c	Tue Jul 17 08:11:18 2007
+++ trunk-new/x264.c	Fri Jan  4 19:46:48 2008
@@ -304,6 +304,7 @@
     H0( "      --no-psnr               Disable PSNR computation\n" );
     H0( "      --no-ssim               Disable SSIM computation\n" );
     H0( "      --threads <integer>     Parallel encoding\n" );
+    H1( "      --thread-queue <integer> Number of delay frames for thread sync\n" );
     H0( "      --thread-input          Run Avisynth in its own thread\n" );
     H1( "      --non-deterministic     Slightly improve quality of SMP, at the cost of repeatability\n" );
     H1( "      --no-asm                Disable all CPU optimizations\n" );
@@ -425,6 +426,7 @@
             { "zones",   required_argument, NULL, 0 },
             { "qpfile",  required_argument, NULL, OPT_QPFILE },
             { "threads", required_argument, NULL, 0 },
+            { "thread-queue", required_argument, NULL, 0 },
             { "thread-input", no_argument,  NULL, OPT_THREAD_INPUT },
             { "non-deterministic", no_argument, NULL, 0 },
             { "no-psnr", no_argument,       NULL, 0 },
diff -uNr trunk/x264.h trunk-new/x264.h
--- trunk/x264.h	Tue Nov 20 02:57:28 2007
+++ trunk-new/x264.h	Fri Jan  4 19:46:48 2008
@@ -142,6 +142,7 @@
     /* CPU flags */
     unsigned int cpu;
     int         i_threads;       /* encode multiple frames in parallel */
+    int         i_thread_queue;  /* number of frames to prepare in advance (>= i_threads) */
     int         b_deterministic; /* whether to allow non-deterministic optimizations when threaded */
 
     /* Video Properties */